
# input: list of char 1,r,s coding a corona
# output: minimal string (amont cyclic/reverse of l) encoding this corona
def canonical(l):
    lr=copy(l)
    lr.reverse()
    z=[[l[mod(j+i,len(l))] for j in range(len(l))] for i in range(len(l))]
    z=z+[[lr[mod(j+i,len(lr))] for j in range(len(lr))] for i in range(len(lr))]
    z.sort()
    x=''
    for i in range(len(l)):
        x=x+z[0][i]
    return x

# input:  string w encoding a corona
# output: corresponding angle vector
def angles(w):
    k=[0,0,0,0,0,0]
    n=len(w)
    for i in range(n):
        if {w[i],w[mod(i+1,n)]}=={'1'}:
            k[0]=k[0]+1
        if {w[i],w[mod(i+1,n)]}=={'1','r'}:
            k[1]=k[1]+1
        if {w[i],w[mod(i+1,n)]}=={'1','s'}:
            k[2]=k[2]+1
        if {w[i],w[mod(i+1,n)]}=={'r'}:
            k[3]=k[3]+1
        if {w[i],w[mod(i+1,n)]}=={'r','s'}:
            k[4]=k[4]+1
        if {w[i],w[mod(i+1,n)]}=={'s'}:
            k[5]=k[5]+1
    return tuple(k)

# input: angle vector k
# output: set of strings encoding all the corresponding coronas
def sequences(k):
    if k[0]+k[1]+k[2]>0:
        return aux_sequences(vector(k),'','1',false)
    elif k[3]+k[1]+k[4]>0:
        return aux_sequences(vector(k),'','r',false)
    else:
        return aux_sequences(vector(k),'','s',false)
        
# same but stop once the first sequence is found
def sequence(k):
    try:
        if k[0]+k[1]+k[2]>0:
            return aux_sequences(vector(k),'','1',true)
        elif k[3]+k[1]+k[4]>0:
            return aux_sequences(vector(k),'','r',true)
        else:
            return aux_sequences(vector(k),'','s',true)
    except Exception as seq:
        return seq.args[0]

# recursive auxiliary function for the two previous ones.
def aux_sequences(k,u,e,une):
    if k==0:
        if une:
            raise Exception(canonical(list(u)))
        return set([canonical(list(u))])
    l=set()
    if e=='1':
        if k[0]>0:
            l.update(aux_sequences(k-vector((1,0,0,0,0,0)),u+'1','1',une))
        if k[1]>0:
            l.update(aux_sequences(k-vector((0,1,0,0,0,0)),u+'r','r',une))
        if k[2]>0:
            l.update(aux_sequences(k-vector((0,0,1,0,0,0)),u+'s','s',une))
    elif e=='r':
        if k[3]>0:
            l.update(aux_sequences(k-vector((0,0,0,1,0,0)),u+'r','r',une))
        if k[1]>0:
            l.update(aux_sequences(k-vector((0,1,0,0,0,0)),u+'1','1',une))
        if k[4]>0:
            l.update(aux_sequences(k-vector((0,0,0,0,1,0)),u+'s','s',une))
    else:
        if k[5]>0:
            l.update(aux_sequences(k-vector((0,0,0,0,0,1)),u+'s','s',une))
        if k[4]>0:
            l.update(aux_sequences(k-vector((0,0,0,0,1,0)),u+'r','r',une))
        if k[2]>0:
            l.update(aux_sequences(k-vector((0,0,1,0,0,0)),u+'1','1',une))
    return l

# check whether the angle vector k does indeed correspond to a corona
def coronable(k):
    # check that there is an integer number of discs in the corona
    a=mod(k[1]+k[2],2)==0 and mod(k[1]+k[4],2)==0 and mod(k[2]+k[4],2)==0
    # if yxy, then either only y's or yxz or xxy
    # si yxy alors soit que des y, soit yxz ou zxy
    b=k[1]<>0 or k[2]<>0 or k[0]==0 or (k[1]==0 and k[2]==0 and k[3]==0 and k[4]==0 and k[5]==0)
    c=k[1]<>0 or k[4]<>0 or k[3]==0 or (k[0]==0 and k[1]==0 and k[2]==0 and k[4]==0 and k[5]==0)
    d=k[2]<>0 or k[4]<>0 or k[5]==0 or (k[0]==0 and k[1]==0 and k[2]==0 and k[3]==0 and k[4]==0)
    # cycles on the triangle between 1,r,s
    z=min(k[1],k[2],k[4])
    e=mod(k[1]-z,2)==0 and mod(k[2]-z,2)==0 and mod(k[4]-z,2)==0
    return a and b and c and d and e

# computes the list of the possible small coronas
def small_coronas():
    pc=[]
    for k in cartesian_product([range(6) for i in range(6)]):
        if coronable(k) and k[0]+k[1]+k[2]+k[3]+k[4]+k[5]<6 and 6<3*k[0]+3*k[1]+3/2*k[2]+3*k[3]+3/2*k[4]+k[5]:
            pc.append(tuple(k))
    return pc

# algebraic ratio s/r given by the small coronas pc[0:9] (which do not contain any 1)
# This corresponds to the values in the paper Ken06
var('x')
kennedy=[
AA(sin(pi/12)/(1-sin(pi/12))),
AA(5-2*sqrt(6)),
AA((2*sqrt(3)+1-2*sqrt(1+sqrt(3)))/3),
AA(ZZ[x](8*x^3+3*x^2-2*x-1).roots(AA)[0][0]),
AA((sqrt(17)-3)/4),
AA(ZZ[x](x^8 - 8*x^7 - 44*x^6 - 232*x^5 - 482*x^4 - 24*x^3 + 388*x^2 - 120*x + 9).roots(AA)[2][0]),
AA(2/sqrt(3)-1),
AA(ZZ[x](x^4-10*x^2-8*x+9).roots(AA)[0][0]),
AA(sqrt(2)-1),
AA(1/sin(pi/5)-1)]

# input: angle vector k of a small corona
# output: list of the medium coronas compatible with the lower bound on s/r given by k
# to have them all: k=pc[1] (minimal lower bound)
def medium_coronas(k):
    # a: lower bound on s/r given by k
    k2=angles(sequence(k).replace('1','r'))
    a=RIF(kennedy[pc.index(k2)])
    # maximal values of the integers l3, l5 and l6
    ua=RIF(arccos(1/(1+a)))
    va=RIF(arccos(1-2*a^2/(1+a)^2))
    dp=RIF(2*pi)
    l3=floor((dp/ua).endpoints()[1])
    l5=l3
    l6=floor((dp/va).endpoints()[1])
    # brute force search among possible angle vectors l
    mc=[]
    for l in cartesian_product([range(6),range(6),range(l3+1),range(6),range(l5+1),range(l6+1)]):
        b=(l[0]+l[1]+l[3]+RIF(3/pi)*(l[2]*ua+l[4]*ua+l[5]*va)).endpoints()[0]
        c=3*l[0]+3/2*l[1]+3/2*l[2]+l[3]+l[4]+l[5]        
        if b<6 and 6<c and l[0]+l[1]+l[3]+(l[2]+l[4])/2<6 and coronable(l):
            mc.append(tuple(l))
    return mc

# compute the angle xyz (not exact because of arccos)
def angle(x,y,z):
    return arccos(((y+x)^2+(y+z)^2-(x+z)^2)/2/(y+x)/(y+z))
# much more precise
angle=lambda x,y,z: arccos(1-2/(y/x+1)/(y/z+1))


# input: values rr and ss for r and s (exact or interval), type tc=0/1/2 of searched coronas
# output: coronas of type tc which are possibly (interval arith.) compatible with rr and ss
def find_coronas_int(rr,ss,tc):
    rrr=RIF(rr)
    sss=RIF(ss)
    un=RIF(1)
    dp=RIF(2*pi)
    z=[sss,rrr,un][tc]
    a=[angle(un,z,un),angle(un,z,rrr),angle(un,z,sss),angle(rrr,z,rrr),angle(rrr,z,sss),angle(sss,z,sss)]
    candidates=[]
    # brute force search, restrict possible values to fasten
    for k0 in range(floor((dp/a[0]).endpoints()[1])+1):
        for k1 in range(floor(((dp-k0*a[0])/a[1]).endpoints()[1])+1):
            for k2 in range(floor(((dp-k0*a[0]-k1*a[1])/a[2]).endpoints()[1])+1):
                for k3 in range (floor(((dp-k0*a[0]-k1*a[1]-k2*a[2])/a[3]).endpoints()[1])+1):
                    for k4 in range (floor(((dp-k0*a[0]-k1*a[1]-k2*a[2]-k3*a[3])/a[4]).endpoints()[1])+1):
                        k5f=((dp-k0*a[0]-k1*a[1]-k2*a[2]-k3*a[3]-k4*a[4])/a[5])
                        for k5 in range(max(0,floor(k5f.endpoints()[0])), floor(k5f.endpoints()[1])+1):
                            if (dp-k0*a[0]-k1*a[1]-k2*a[2]-k3*a[3]-k4*a[4]-k5*a[5]).contains_zero():
                                k=tuple((k0,k1,k2,k3,k4,k5))
                                # at most six discs of size x around a disc of size x
                                if (tc==0 and k5+(k2+k4)/2<=6) or (tc==1 and k3+(k1+k4)/2<=6) or (tc==2 and k0+(k1+k2)/2<=6):
                                    if coronable(k):
                                        candidates.append(k)
    # removes hexagonal compact
    candidates.remove([(0,0,0,0,0,6),(0,0,0,6,0,0),(6,0,0,0,0,0)][tc])
    return candidates

# add verification holes if interstitial ?


# input: angle vector k of a corona of type tc, values rr and ss for r and s
# ouput: is the corona possibly (interval arithmetic) compatible?
def check_corona_int(k,tc,rr,ss):
    rrr=RIF(rr)
    sss=RIF(ss)
    un=RIF(1)
    dp=RIF(2*pi)
    z=[sss,rrr,un][tc]
    e=k[0]*angle(un,z,un)+k[1]*angle(un,z,rrr)+k[2]*angle(un,z,sss)+k[3]*angle(rrr,z,rrr)+k[4]*angle(rrr,z,sss)+k[5]*angle(sss,z,sss)-dp
    return e.contains_zero()

